<?php

namespace Bolt;

use Silex;
use Symfony\Component\HttpFoundation\Response;

/**
 * Wrapper around Twig's render() function. Handles the following responsibilities:
 *
 * - Calls twig's render
 * - Stores a page in cache, if needed
 * - Store template (partials) in cache, if needed
 * - Fetches pages or template (partials) from cache
 *
 * @author Bob den Otter, bob@twokings.nl
 *
 */
class Render
{

    /**
     * Set up the object.
     *
     * @param Silex\Application $app
     */
    public function __construct(Application $app)
    {
        $this->app = $app;
    }

    /**
     *
     * Render a template, possibly store it in cache. Or, if applicable, return the cached result
     *
     * @param $template
     * @param array $vars
     * @return mixed
     */
    public function render($template, $vars = array())
    {

        // Start the 'stopwatch' for the profiler.
        $this->app['stopwatch']->start('bolt.render', 'template');

        if ($html = $this->fetchCachedTemplate($template)) {

            // Do nothing.. The page is fetched from cache..

        } else {

            $html = $this->app['twig']->render($template, $vars);
            $this->cacheRenderedTemplate($template, $html);

        }

        // Stop the 'stopwatch' for the profiler.
        $this->app['stopwatch']->stop('bolt.render');

        return $html;

    }

    /**
     * Postprocess the rendered HTML: insert the snippets, and stuff.
     *
     * @param Response $response
     * @return string
     */
    public function postProcess(Response $response)
    {
        $html = $response->getContent();
        $html = $this->app['extensions']->processSnippetQueue($html);
        $this->cacheRequest($html);

        return $html;

    }


    /**
     * Retrieve a  page (or basically, any template) from cache
     *
     * @param string $template
     * @return mixed
     */
    public function fetchCachedTemplate($template)
    {
        if ($this->checkCacheConditions('templates', true)) {
            $key = md5($template . $this->app['request']->getRequestUri());

            return $this->app['cache']->fetch($key);

        }
    }

    /**
     * Retrieve a fully cached page from cache.
     *
     * @return mixed
     */
    public function fetchCachedRequest()
    {
        if ($this->checkCacheConditions('request', true)) {
            $key = md5($this->app['request']->getPathInfo());

            return $this->app['cache']->fetch($key);

        }
    }

    /**
     * Store a page (or basically, any template) to cache.
     *
     * @param $template
     * @param $html
     */
    public function cacheRenderedTemplate($template, $html)
    {

        if ($this->checkCacheConditions('templates')) {

            // Store it part-wise, with the correct template name..
            $key = md5($template . $this->app['request']->getRequestUri());
            $this->app['cache']->save($key, $html, $this->cacheDuration());

        }

    }

    /**
     * Store a fully rendered (and postprocessed) page to cache.
     *
     * @param $html
     */
    public function cacheRequest($html)
    {

        if ($this->checkCacheConditions('request')) {

            // This is where the magic happens.. We also store it with an empty 'template' name,
            // So we can later fetch it by its request..
            $key = md5($this->app['request']->getPathInfo());
            $this->app['cache']->save($key, $html, $this->cacheDuration());

        }

    }

    /**
     * Get the duration (in seconds) for the cache.
     *
     * @return int;
     */
    public function cacheDuration()
    {
        // in minutes..
        $duration = $this->app['config']->get('general/caching/duration', 10);

        // in seconds.
        return $duration * 60;

    }

    /**
     * Check if the current conditions are suitable for caching.
     *
     * @param string $type
     * @param bool $checkoverride
     * @return bool
     */
    public function checkCacheConditions($type = 'template', $checkoverride = false)
    {

        // Only cache pages in the frontend.
        if ($this->app['end'] != "frontend") {
            return false;
        }

        // Only cache for 'get' requests.
        if ($this->app['request']->getMethod() != "GET") {
            return false;
        }

        // Don't use the cache, if not enabled in the config.
        if (!$this->app['config']->get('general/caching/' . $type)) {
            return false;
        }

        // Don't use the cache, if we're currently logged in. (unless explicitly enabled in config.yml
        if (!$this->app['config']->get('general/caching/authenticated') &&
            $this->app['users']->getCurrentUsername() != "") {
            return false;
        }

        // if we've added 'force_refresh=1', we don't use the cache. Note, in most cases,
        // we don't _fetch_ from the cache, but we do allow _saving_ to the cache.
        if ($checkoverride && $this->app['request']->get('force_refresh') == 1) {
            return false;
        }

        // All's well!
        return true;

    }
}
